"""Spider client TaskGraph module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from spider_py import core
from spider_py.client.task import create_task, TaskFunction

if TYPE_CHECKING:
    from collections.abc import Sequence


class TaskGraph:
    """
    Represents a client-side task graph.

    This class is a wrapper of `spider_py.core.Task`.
    """

    def __init__(self) -> None:
        """Initializes TaskGraph."""
        self._impl = core.TaskGraph()


def group(tasks: Sequence[TaskFunction | TaskGraph]) -> TaskGraph:
    """
    Groups task functions and task graph into a single task graph.
    :param tasks: List of task functions or task graphs.
    :return: The new task graph.
    """
    graph = TaskGraph()
    for task in tasks:
        if callable(task):
            graph._impl.add_task(create_task(task))
        else:
            graph._impl.merge_graph(task._impl)

    return graph


def chain(parent: TaskFunction | TaskGraph, child: TaskFunction | TaskGraph) -> TaskGraph:
    """
    Chains two task functions or task graphs into a single task graph.
    :param parent:
    :param child:
    :return: The new task graph.
    :raises TypeError: If the parent outputs and child inputs do not match.
    """
    parent_core_graph: core.TaskGraph
    child_core_graph: core.TaskGraph

    if callable(parent):
        parent_core_graph = core.TaskGraph()
        parent_core_graph.add_task(create_task(parent))
    else:
        parent_core_graph = parent._impl

    if callable(child):
        child_core_graph = core.TaskGraph()
        child_core_graph.add_task(create_task(child))
    else:
        child_core_graph = child._impl

    graph = TaskGraph()
    graph._impl = core.TaskGraph.chain_graph(parent_core_graph, child_core_graph)
    return graph



"""Task executor message module."""

from enum import IntEnum
from typing import cast

import msgpack


class TaskExecutorResponseType(IntEnum):
    """Task executor response type."""

    Unknown = 0
    Result = 1
    Error = 2
    Block = 3
    Ready = 4
    Cancel = 5


class TaskExecutorRequestType(IntEnum):
    """Task executor request type."""

    Unknown = 0
    Arguments = 1
    Resume = 2


class InvalidRequestTypeError(Exception):
    """Exception raised for invalid request types."""

    def __init__(self, message: str) -> None:
        """Initializes the InvalidRequestTypeError with a message."""
        super().__init__(message)


ArgRequestLength = 2


def get_request_body(message: bytes) -> list[object]:
    """
    Gets the request body from the request message.
    :param message: The msgpack serialized request message.
    :return: The request body as a list of objects.
    :raises TypeError: If the serialized message is not a msgpack list or the list length is
        incorrect.
    :raises InvalidRequestTypeError: If the message header is not
        `TaskExecutorRequestType.Arguments`.
    """
    data = msgpack.unpackb(message)
    if not isinstance(data, list):
        msg = "Message is not a list."
        raise TypeError(msg)
    if len(data) != ArgRequestLength:
        msg = f"Message is not a list with {ArgRequestLength} elements, got {len(data)}."
        raise TypeError(msg)
    header = data[0]
    if not isinstance(header, int):
        msg = "Message header is not an int."
        raise TypeError(msg)
    if TaskExecutorRequestType.Arguments != header:
        msg = f"Message header is not an `Arguments`: {header}."
        raise InvalidRequestTypeError(msg)
    body = data[1]
    if not isinstance(body, list):
        msg = "Message body is not a list."
        raise TypeError(msg)
    return cast("list[object]", body)



"""Task module for Spider."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from uuid import UUID, uuid4

from spider_py.core.data import Data, DataId

TaskId = UUID


@dataclass
class TaskInputOutput:
    """Represents a task input that references the output of another task by its ID and position."""

    task_id: TaskId
    position: int


TaskInputValue = bytes
TaskInputData = Data | DataId


@dataclass
class TaskInput:
    """Represents a task input"""

    type: str
    value: TaskInputData | TaskInputOutput | TaskInputValue | None


TaskOutputValue = bytes
TaskOutputData = Data | DataId


@dataclass
class TaskOutput:
    """Represents a task output"""

    type: str
    value: TaskOutputData | TaskOutputValue


class TaskState(IntEnum):
    """Represents the state of a task"""

    Pending = 0
    Ready = 1
    Running = 2
    Succeeded = 3
    Failed = 4
    Cancelled = 5

    def __str__(self) -> str:
        """
        Returns string representation of task state.
        :return: The string representation of task state.
        """
        return _StateToStrMap[self]

    @staticmethod
    def from_str(state_str: str) -> TaskState:
        """
        Returns task state from string representation.
        :param state_str: The string representation of task state.
        :return: The task state from string representation.
        :raises ValueError: If the state string is not recognized.
        """
        state = _StrToStateMap.get(state_str)
        if state is not None:
            return state
        msg = f"Invalid task state string: {state_str}"
        raise ValueError(msg)


_StateToStrMap = {
    TaskState.Pending: "pending",
    TaskState.Ready: "ready",
    TaskState.Running: "running",
    TaskState.Succeeded: "success",
    TaskState.Failed: "fail",
    TaskState.Cancelled: "cancel",
}

_StrToStateMap = {value: key for key, value in _StateToStrMap.items()}


@dataclass
class Task:
    """Represents a task in Spider."""

    task_id: TaskId = field(default_factory=uuid4)
    function_name: str = ""
    state: TaskState = TaskState.Pending
    timeout: float = 0
    max_retries: int = 0
    task_inputs: list[TaskInput] = field(default_factory=list)
    task_outputs: list[TaskOutput] = field(default_factory=list)

    def set_pending(self) -> None:
        """Sets the task state to Pending."""
        self.state = TaskState.Pending

    def set_ready(self) -> None:
        """Sets the task state to Ready."""
        self.state = TaskState.Ready



"""Spider client task module."""

import inspect
from collections.abc import Callable
from types import FunctionType, GenericAlias
from typing import get_args, get_origin

from spider_py import core
from spider_py.client.data import Data
from spider_py.client.task_context import TaskContext
from spider_py.core import DataId, TaskInput, TaskOutput, TaskOutputValue
from spider_py.type import to_tdl_type_str

# NOTE: This type alias is for clarification purposes only. It does not enforce static type checks.
# Instead, we rely on the runtime check to ensure the first argument is `TaskContext`. To statically
# enforce the first argument to be `TaskContext`, `Protocol` is required, which is not compatible
# with `Callable` without explicit type casting.
TaskFunction = Callable[..., object]


def _is_tuple(t: type | GenericAlias) -> bool:
    """
    :param t:
    :return: Whether t is a tuple.
    """
    return get_origin(t) is tuple


def _validate_and_convert_params(signature: inspect.Signature) -> list[TaskInput]:
    """
    Validates the task parameters and converts them into a list of `core.TaskInput`.
    :param signature:
    :return: The converted task parameters.
    :raises TypeError: If the parameters are invalid.
    """
    params = list(signature.parameters.values())
    inputs = []
    if not params or params[0].annotation is not TaskContext:
        msg = "First argument is not a TaskContext."
        raise TypeError(msg)
    for param in params[1:]:
        if param.kind in {inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD}:
            msg = "Variadic parameters are not supported."
            raise TypeError(msg)
        if param.annotation is inspect.Parameter.empty:
            msg = "Parameters must have type annotation."
            raise TypeError(msg)
        tdl_type_str = to_tdl_type_str(param.annotation)
        inputs.append(TaskInput(tdl_type_str, None))
    return inputs


def _validate_and_convert_return(signature: inspect.Signature) -> list[TaskOutput]:
    """
    Validates the task returns and converts them into a list of `core.TaskOutput`.
    :param signature:
    :return: The converted task returns.
    :raises TypeError: If the return type is invalid.
    """
    returns = signature.return_annotation
    outputs = []
    if returns is inspect.Parameter.empty:
        msg = "Return type must have type annotation."
        raise TypeError(msg)

    if not _is_tuple(returns):
        tdl_type_str = to_tdl_type_str(returns)
        if returns is Data:
            outputs.append(TaskOutput(tdl_type_str, DataId(int=0)))
        else:
            outputs.append(TaskOutput(tdl_type_str, TaskOutputValue()))
        return outputs

    args = get_args(returns)
    if Ellipsis in args:
        msg = "Variable-length tuple return types are not supported."
        raise TypeError(msg)
    for arg in args:
        tdl_type_str = to_tdl_type_str(arg)
        if arg is Data:
            outputs.append(TaskOutput(tdl_type_str, DataId(int=0)))
        else:
            outputs.append(TaskOutput(tdl_type_str, TaskOutputValue()))
    return outputs


def create_task(func: TaskFunction) -> core.Task:
    """
    Creates a core Task object from the task function.
    :param func:
    :return: The created core Task object.
    :raise TypeError: If the function signature contains unsupported types.
    """
    if not isinstance(func, FunctionType):
        msg = "`func` is not a function."
        raise TypeError(msg)
    signature = inspect.signature(func)
    return core.Task(
        function_name=f"{func.__module__}.{func.__qualname__}",
        task_inputs=_validate_and_convert_params(signature),
        task_outputs=_validate_and_convert_return(signature),
    )



"""Spider package root."""

from spider_py.client import chain, Data, Driver, group, Job, TaskContext, TaskGraph
from spider_py.core import JobStatus
from spider_py.type import Double, Float, Int8, Int16, Int32, Int64

__all__ = [
    "Data",
    "Double",
    "Driver",
    "Float",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Job",
    "JobStatus",
    "TaskContext",
    "TaskGraph",
    "chain",
    "group",
]



"""Serialization and deserialization into serializable format."""

from dataclasses import fields, is_dataclass
from types import GenericAlias
from typing import Any, cast, get_args, get_origin, get_type_hints


def to_serializable(obj: object, cls: type | GenericAlias) -> object:
    """
    Converts an object into a serializable format consisting only of built-in primitive types and
    collections (lists and dictionaries).

    - Built-in container types (lists or dictionaries) are recursively transformed.
    - Dataclass instances are converted into dictionaries mapping field names to their serialized
      values.
    - All other objects are returned as-is.

    :param obj: Object to serialize. Must be of types supported by Spider TDL.
    :param cls: The expected type of `obj`. Must be a concrete type or GenericAlias supported by
        Spider TDL.
    :return: A serializable representation of `obj`.
    :raise: TypeError if `obj`'s type does not match `cls`.
    :raise: TypeError if `cls` is not a TDL supported type.
    """
    origin = get_origin(cls)
    if origin is None:
        expected_type = cast("type", cls)
        if not isinstance(obj, expected_type):
            msg = f"Object {obj!r} is not of type {expected_type!r}."
            raise TypeError(msg)
        if is_dataclass(expected_type):
            return _to_serializable_dataclass(obj, expected_type)
        return obj

    if not isinstance(obj, origin):
        msg = f"Object {obj!r} is not of type {cls!r}."
        raise TypeError(msg)

    if origin is list:
        (element_type,) = get_args(cast("GenericAlias", cls))
        return _to_serializable_list(obj, element_type)

    if origin is dict:
        (key_type, value_type) = get_args(cast("GenericAlias", cls))
        return _to_serializable_dict(obj, key_type, value_type)

    msg = f"Unsupported type: {cls!r}."
    raise TypeError(msg)


def from_serializable(cls: type | GenericAlias, data: object) -> object:
    """
    Deserializes data in a serializable format back into an instance of the specified `cls`.

    :param cls: Class to deserialize into. Must be a concrete type, list, or dict.
    :param data: Data in serializable form.
    :return: An instance of `cls` reconstructed from the serialized data.
    :raise: TypeError if `data` is not compatible with `cls`.
    """
    msg = f"Cannot create an instance of {cls} with {data!r}."

    origin = get_origin(cls)
    if origin is None:
        # If `cls` does not have an origin, it is a concrete type.
        return _deserialize_as_class(cast("type", cls), data)

    if origin is list:
        (key_type,) = get_args(cls)
        if not isinstance(data, list):
            raise TypeError(msg)
        return [from_serializable(key_type, item) for item in data]

    if origin is dict:
        key_type, value_type = get_args(cls)
        if not isinstance(data, dict):
            raise TypeError(msg)
        return {
            from_serializable(key_type, key): from_serializable(value_type, value)
            for key, value in data.items()
        }

    raise TypeError(msg)


def _deserialize_as_class(cls: type, data: object) -> object:
    """
    Deserializes the input data as a `cls` instance.
    :param cls: Class to deserialize into. Must be a non-container type.
    :param data: Serialized data.
    :return: An instance of `cls` reconstructed from the serialized data.
    :raise: TypeError if `data` is not compatible with `cls`.
    """
    msg = f"Cannot create an instance of {cls} with {data!r}."
    if not is_dataclass(cls):
        # Fall back to normal constructor if `cls` is not a dataclass.
        return cls(data)

    if not isinstance(data, dict):
        raise TypeError(msg)
    args = {}
    expected_fields = {field.name: field for field in fields(cls)}
    for name, value in data.items():
        if name not in expected_fields:
            raise TypeError(msg)
        expected_field_type = get_type_hints(cls).get(name)
        if not isinstance(expected_field_type, (type, GenericAlias)):
            raise TypeError(msg)
        args[name] = from_serializable(expected_field_type, value)
    return cls(**args)


def _to_serializable_dataclass(obj: object, cls: type) -> object:
    """
    Converts a dataclass to a serializable format.

    :param obj: Dataclass object to serialize.
    :param cls: The expected type of `obj`.
    :return: A serializable representation of `obj`.
    :raise: TypeError if any field in `cls` is of an unsupported type.
    """
    if not is_dataclass(obj):
        # Control flow should not reach here. However, this check is required to silence ruff.
        msg = f"Object {obj!r} is not a dataclass instance."
        raise TypeError(msg)
    serialized_dict = {}
    for field in fields(obj):
        field_value = getattr(obj, field.name)
        field_type = get_type_hints(cls).get(field.name, None)
        if field_type is None or not isinstance(field_type, (type, GenericAlias)):
            msg = f"Invalid field type for {field.name} in {cls!r}."
            raise TypeError(msg)
        serialized_value = to_serializable(field_value, field_type)
        serialized_dict[field.name] = serialized_value
    return serialized_dict


def _to_serializable_list(obj: list[Any], element_type: type | GenericAlias) -> object:
    """
    Converts a list to a serializable format.

    :param obj: List object to serialize.
    :param element_type: Type of elements in the list.
    :return: A serializable representation of the list.
    """
    serialized_list = []
    for item in obj:
        serialized_item = to_serializable(item, element_type)
        serialized_list.append(serialized_item)
    return serialized_list


def _to_serializable_dict(
    obj: dict[Any, Any], key_type: type | GenericAlias, value_type: type | GenericAlias
) -> object:
    """
    Converts a dictionary to a serializable format.

    :param obj: Dictionary object to serialize.
    :param key_type: Type of keys in the dictionary.
    :param value_type: Type of values in the dictionary.
    :return: A serializable representation of the dictionary.
    """
    serialized_dict = {}
    for key, value in obj.items():
        serialized_key = to_serializable(key, key_type)
        serialized_value = to_serializable(value, value_type)
        serialized_dict[serialized_key] = serialized_value
    return serialized_dict



"""Spider job module."""

import msgpack

from spider_py import core
from spider_py.client.data import Data
from spider_py.storage import Storage, StorageError
from spider_py.storage.job_utils import fetch_and_update_job_results, fetch_and_update_job_status
from spider_py.type import parse_tdl_type
from spider_py.utils import from_serializable


class Job:
    """Represents Spider job."""

    def __init__(self, job: core.Job, storage: Storage) -> None:
        """
        Creates a new Spider job.
        :param job:
        :param storage: The storage backend.
        """
        self._impl = job
        self._storage = storage

    def get_status(self) -> core.JobStatus:
        """:return: The current job status."""
        if self._impl.is_running():
            fetch_and_update_job_status(self._storage, self._impl)
        return self._impl.status

    def get_results(self) -> tuple[object, ...] | None:
        """
        :return: The job results if the job ended successfully.
        :return: None if the job is still running or ended unsuccessfully.
        """
        fetch_and_update_job_results(self._storage, self._impl)

        if self._impl.results is None:
            return None

        return _deserialize_outputs(self._impl.results)


def _deserialize_outputs(outputs: list[core.TaskOutput]) -> tuple[object, ...]:
    """
    Deserializes a list of `core.TaskOutput` objects into their corresponding Python values.
    :param outputs:
    :return: A tuple of deserialized values.
    """
    results = []
    for output in outputs:
        if isinstance(output.value, core.TaskOutputValue):
            type_name = output.type
            if isinstance(type_name, bytes):
                type_name = type_name.decode("utf-8")
            cls = parse_tdl_type(type_name).native_type()
            unpacked = msgpack.unpackb(output.value, raw=False, strict_map_key=False)
            results.append(from_serializable(cls, unpacked))
        elif isinstance(output.value, core.Data):
            results.append(Data(output.value))
        else:
            msg = "Unsupported output type."
            raise StorageError(msg)
    return tuple(results)



"""MariaDB Storage module."""

from collections.abc import Sequence
from uuid import UUID, uuid4

import mariadb
from typing_extensions import override

from spider_py import core
from spider_py.storage.jdbc_url import JdbcParameters
from spider_py.storage.storage import Storage, StorageError

InsertJob = """
INSERT INTO
  `jobs` (`id`, `client_id`)
VALUES
  (?, ?)"""

InsertTask = """
INSERT INTO
  `tasks` (`id`, `job_id`, `func_name`, `language`, `state`, `timeout`, `max_retry`)
VALUES
  (?, ?, ?, ?, ?, ?, ?)"""

InsertTaskInputOutput = """
INSERT INTO
  `task_inputs` (`task_id`, `position`, `type`, `output_task_id`, `output_task_position`)
VALUES
  (?, ?, ?, ?, ?)"""

InsertTaskInputData = """
INSERT INTO
  `task_inputs` (`task_id`, `position`, `type`, `data_id`)
VALUES
  (?, ?, ?, ?)"""

InsertTaskInputValue = """
INSERT INTO
  `task_inputs` (`task_id`, `position`, `type`, `value`)
VALUES
  (?, ?, ?, ?)"""

InsertTaskOutput = """
INSERT INTO
  `task_outputs` (`task_id`, `position`, `type`)
VALUES
  (?, ?, ?)"""

InsertTaskDependency = """
INSERT INTO
  `task_dependencies` (parent, child)
VALUES
  (?, ?)"""

InsertInputTask = """
INSERT INTO
  `input_tasks` (`job_id`, `task_id`, `position`)
VALUES
  (?, ?, ?)"""

InsertOutputTask = """
INSERT INTO
  `output_tasks` (`job_id`, `task_id`, `position`)
VALUES
  (?, ?, ?)"""


GetJobStatus = """
SELECT
  `state`
FROM
  `jobs`
WHERE
  `id` = ?"""

GetOutputTasks = """
SELECT
  `task_id`
FROM
  `output_tasks`
WHERE
  `job_id` = ?
ORDER BY
  `position`"""

GetTaskOutputs = """
SELECT
  `type`,
  `value`,
  `data_id`
FROM
  `task_outputs`
WHERE
  `task_id` = ?
ORDER BY
  `position`"""

InsertData = """
INSERT INTO
  `data` (`id`, `value`, `hard_locality`)
VALUES
  (?, ?, ?)"""

InsertDataLocality = """
INSERT INTO
  `data_locality` (`id`, `address`)
VALUES
  (?, ?)"""

InsertDataRefDriver = """
INSERT INTO
  `data_ref_driver` (`id`, `driver_id`)
VALUES
  (?, ?)"""

InsertDataRefTask = """
INSERT INTO
  `data_ref_task` (`id`, `task_id`)
VALUES
  (?, ?)"""

GetData = """
SELECT
  `value`,
  `hard_locality`
FROM
  `data`
WHERE
  `id` = ?"""

GetDataLocality = """
SELECT
  `address`
FROM
  `data_locality`
WHERE
  `id` = ?"""

InsertDriver = """
INSERT INTO
  `drivers` (`id`)
VALUES
  (?)"""

_StrToJobStatusMap = {
    "running": core.JobStatus.Running,
    "success": core.JobStatus.Succeeded,
    "fail": core.JobStatus.Failed,
    "cancel": core.JobStatus.Cancelled,
}


class MariaDBStorage(Storage):
    """MariaDB Storage class."""

    def __del__(self) -> None:
        """
        Closes the connection to the MariaDB database.
        :raises StorageError: If closing the connection fails.
        """
        self._conn.close()

    def __init__(self, params: JdbcParameters) -> None:
        """
        Connects to the MariaDB database.
        :param params: The JDBC parameters for connecting to the database.
        :raises StorageError: If the connection to the database fails.
        """
        try:
            self._conn = mariadb.connect(
                host=params.host,
                port=params.port,
                user=params.user,
                password=params.password,
                database=params.database,
            )
        except mariadb.Error as e:
            raise StorageError(str(e)) from e

    @override
    def submit_jobs(
        self, driver_id: core.DriverId, task_graphs: Sequence[core.TaskGraph]
    ) -> Sequence[core.Job]:
        if not task_graphs:
            return []
        try:
            # Create job UUIDs and task UUIDs
            jobs = []
            task_ids = []
            for task_graph in task_graphs:
                jobs.append(core.Job(uuid4()))
                task_ids.append([uuid4() for _ in task_graph.tasks])

            with self._conn.cursor() as cursor:
                # Insert jobs table
                cursor.executemany(InsertJob, [(job.job_id.bytes, driver_id.bytes) for job in jobs])
                # Insert tasks table
                cursor.executemany(
                    InsertTask,
                    self._gen_task_insertion_params(jobs, task_ids, task_graphs),
                )

                # Insert task dependencies table
                dep_params = self._gen_task_dependency_insertion_params(task_ids, task_graphs)
                if dep_params:
                    cursor.executemany(
                        InsertTaskDependency,
                        dep_params,
                    )

                # Insert input tasks table
                cursor.executemany(
                    InsertInputTask,
                    self._gen_input_task_insertion_params(jobs, task_ids, task_graphs),
                )

                # Insert output tasks table
                cursor.executemany(
                    InsertOutputTask,
                    self._gen_output_task_insertion_params(jobs, task_ids, task_graphs),
                )

                # Insert task outputs table
                cursor.executemany(
                    InsertTaskOutput,
                    self._gen_task_output_insertion_params(task_ids, task_graphs),
                )

                # Insert task input data table
                input_data_params = self._gen_task_input_data_insertion_params(
                    task_ids, task_graphs
                )
                if input_data_params:
                    cursor.executemany(
                        InsertTaskInputData,
                        input_data_params,
                    )

                # Insert task input values table
                input_value_params = self._gen_task_input_value_insertion_params(
                    task_ids, task_graphs
                )
                if input_value_params:
                    cursor.executemany(
                        InsertTaskInputValue,
                        input_value_params,
                    )

                # Insert task input outputs table
                input_output_params = self._gen_task_input_output_ref_insertion_params(
                    task_ids, task_graphs
                )
                if input_output_params:
                    cursor.executemany(
                        InsertTaskInputOutput,
                        input_output_params,
                    )

                self._conn.commit()
                return jobs
        except mariadb.Error as e:
            self._conn.rollback()
            raise StorageError(str(e)) from e

    @override
    def get_job_status(self, job: core.Job) -> core.JobStatus:
        try:
            with self._conn.cursor() as cursor:
                status = self._get_job_status(cursor, job)
                self._conn.commit()
                return status
        except mariadb.Error as e:
            self._conn.rollback()
            raise StorageError(str(e)) from e
        except StorageError:
            self._conn.rollback()
            raise

    @override
    def get_job_results(self, job: core.Job) -> list[core.TaskOutput] | None:
        try:
            with self._conn.cursor() as cursor:
                status = self._get_job_status(cursor, job)
                if status != core.JobStatus.Succeeded:
                    self._conn.commit()
                    return None

                cursor.execute(GetOutputTasks, (job.job_id.bytes,))
                task_ids = [task_id for (task_id,) in cursor.fetchall()]

                results = []
                for task_id in task_ids:
                    cursor.execute(GetTaskOutputs, (task_id,))
                    for output_type, value, data_id in cursor.fetchall():
                        if value is not None:
                            results.append(
                                core.TaskOutput(
                                    type=output_type,
                                    value=core.TaskOutputValue(value),
                                )
                            )
                        elif data_id is not None:
                            data = self._get_data(cursor, core.DataId(bytes=data_id))
                            results.append(
                                core.TaskOutput(
                                    type=output_type,
                                    value=data,
                                )
                            )
                        else:
                            msg = "Invalid task output"
                            _raise_storage_error(msg)
                self._conn.commit()
                return results
        except mariadb.Error as e:
            self._conn.rollback()
            raise StorageError(str(e)) from e
        except StorageError:
            self._conn.rollback()
            raise

    @override
    def create_data_with_driver_ref(self, driver_id: core.DriverId, data: core.Data) -> None:
        self._create_data_with_ref(data, InsertDataRefDriver, driver_id)

    @override
    def create_data_with_task_ref(self, task_id: core.TaskId, data: core.Data) -> None:
        self._create_data_with_ref(data, InsertDataRefTask, task_id)

    @override
    def get_data(self, data_id: core.DataId) -> core.Data:
        try:
            with self._conn.cursor() as cursor:
                data = self._get_data(cursor, data_id)
                self._conn.commit()
                return data
        except mariadb.Error as e:
            self._conn.rollback()
            raise StorageError(str(e)) from e

    @override
    def create_driver(self, driver_id: core.DriverId) -> None:
        try:
            with self._conn.cursor() as cursor:
                cursor.execute(InsertDriver, (driver_id.bytes,))
                self._conn.commit()
        except mariadb.Error as e:
            self._conn.rollback()
            raise StorageError(str(e)) from e

    def _create_data_with_ref(
        self, data: core.Data, insert_stmt: str, ref_id: core.DriverId | core.TaskId
    ) -> None:
        """
        Creates a data object in the storage with the given `ref_id` references to the data.
        :param data: The data object to create.
        :param insert_stmt: The SQL statement to insert the reference.
        :param ref_id: The reference ID.
        :raises StorageError: If the storage operations fail.
        """
        try:
            with self._conn.cursor() as cursor:
                cursor.execute(
                    InsertData,
                    (data.id.bytes, data.value, data.hard_locality),
                )
                if data.localities:
                    cursor.executemany(
                        InsertDataLocality,
                        [(data.id.bytes, addr) for addr in data.localities],
                    )
                cursor.execute(
                    insert_stmt,
                    (data.id.bytes, ref_id.bytes),
                )
                self._conn.commit()
        except mariadb.Error as e:
            self._conn.rollback()
            raise StorageError(str(e)) from e

    @staticmethod
    def _gen_task_insertion_params(
        jobs: Sequence[core.Job],
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, bytes, str, str, str, float, int]]:
        """
        Generates parameters for inserting tasks into the database.
        :param jobs: The jobs.
        :param task_ids: The task IDs. Must be the same length as `jobs`.
        :param task_graphs: The task graphs. Must be the same length as `jobs`.
        :return: A list of tuples containing the parameters for each task. Each tuple contains:
            - Task ID.
            - Job ID.
            - Task function name.
            - Task language (always "python").
            - Task state.
            - Task timeout.
            - Task max retry.
        :raises ValueError: If the lengths of `jobs` and `task_graphs` do not match.
        """
        task_insert_params = []
        if len(jobs) != len(task_graphs):
            msg = "The lengths of `jobs` and `task_graphs` must match."
            raise ValueError(msg)
        for graph_index, (job, task_graph) in enumerate(zip(jobs, task_graphs, strict=True)):
            for task_index, task in enumerate(task_graph.tasks):
                task_insert_params.append(
                    (
                        task_ids[graph_index][task_index].bytes,
                        job.job_id.bytes,
                        task.function_name,
                        "python",
                        str(task.state),
                        task.timeout,
                        task.max_retries,
                    )
                )
        return task_insert_params

    @staticmethod
    def _gen_task_dependency_insertion_params(
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, bytes]]:
        """
        Generates parameters for inserting task dependencies into the database.
        :param task_ids: The task IDs.
        :param task_graphs: The task graphs. Must be the same length as `task_ids`.
        :return: A list of tuples containing the parameters for each task dependency. Each tuple
            contains:
            - Parent task ID.
            - Child task ID.
        """
        dep_params = []
        for graph_index, task_graph in enumerate(task_graphs):
            for parent, child in task_graph.dependencies:
                dep_params.append(
                    (task_ids[graph_index][parent].bytes, task_ids[graph_index][child].bytes)
                )
        return dep_params

    @staticmethod
    def _gen_input_task_insertion_params(
        jobs: Sequence[core.Job],
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, bytes, int]]:
        """
        Generates parameters for inserting input tasks into the database.
        :param jobs: The jobs.
        :param task_ids: The task IDs. Must be the same length as `jobs`.
        :param task_graphs: The task graphs. Must be the same length as `jobs`.
        :return: A list of tuples containing the parameters for each input task. Each tuple
            contains:
            - Job ID.
            - Task ID.
            - The positional index of the input task.
        :raises ValueError: If the lengths of `jobs` and `task_graphs` do not match.
        """
        input_task_params = []
        if len(jobs) != len(task_graphs):
            msg = "The lengths of `jobs` and `task_graphs` must match."
            raise ValueError(msg)
        for graph_index, (job, task_graph) in enumerate(zip(jobs, task_graphs, strict=True)):
            for position, task_index in enumerate(task_graph.input_task_indices):
                input_task_params.append(
                    (job.job_id.bytes, task_ids[graph_index][task_index].bytes, position)
                )
        return input_task_params

    @staticmethod
    def _gen_output_task_insertion_params(
        jobs: Sequence[core.Job],
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, bytes, int]]:
        """
        Generates parameters for inserting output tasks into the database.
        :param jobs: The jobs.
        :param task_ids: The task IDs. Must be the same length as `jobs`.
        :param task_graphs: The task graphs. Must be the same length as `jobs`.
        :return: A list of tuples containing the parameters for each output task. Each tuple
            contains:
            - Job ID.
            - Task ID.
            - The positional index of the output task.
        :raises ValueError: If the lengths of `jobs` and `task_graphs` do not match.
        """
        output_task_params = []
        if len(jobs) != len(task_graphs):
            msg = "The lengths of `jobs` and `task_graphs` must match."
            raise ValueError(msg)
        for graph_index, (job, task_graph) in enumerate(zip(jobs, task_graphs, strict=True)):
            for position, task_index in enumerate(task_graph.output_task_indices):
                output_task_params.append(
                    (job.job_id.bytes, task_ids[graph_index][task_index].bytes, position)
                )
        return output_task_params

    @staticmethod
    def _gen_task_output_insertion_params(
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, int, str]]:
        """
        Generates parameters for inserting task outputs into the database.
        :param task_ids: The task IDs.
        :param task_graphs: The task graphs. Must be the same length as `task_ids`.
        :return: A list of tuples containing the parameters for each task output. Each tuple
            contains:
            - Task ID.
            - Positional index of the output.
            - Type of the output.
        """
        output_params = []
        for graph_index, task_graph in enumerate(task_graphs):
            for task_index, task in enumerate(task_graph.tasks):
                for position, task_output in enumerate(task.task_outputs):
                    output_params.append(
                        (task_ids[graph_index][task_index].bytes, position, task_output.type)
                    )
        return output_params

    @staticmethod
    def _gen_task_input_data_insertion_params(
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, int, str, bytes]]:
        """
        Generates parameters for inserting task input data into the database.
        :param task_ids: The task IDs.
        :param task_graphs: The task graphs. Must be the same length as `task_ids`.
        :return: A list of tuples containing the parameters for each task input data. Each tuple
            contains:
            - Task ID.
            - Positional index of the input.
            - Type of the input.
            - Input data.
        """
        input_data_params = []
        for graph_index, task_graph in enumerate(task_graphs):
            for task_index, task in enumerate(task_graph.tasks):
                for position, task_input in enumerate(task.task_inputs):
                    if not isinstance(task_input.value, core.TaskInputData):
                        continue
                    value = task_input.value
                    data = value.id.bytes if isinstance(value, core.Data) else value.bytes
                    input_data_params.append(
                        (
                            task_ids[graph_index][task_index].bytes,
                            position,
                            task_input.type,
                            data,
                        )
                    )
        return input_data_params

    @staticmethod
    def _gen_task_input_value_insertion_params(
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, int, str, bytes]]:
        """
        Generates parameters for inserting task input values into the database.
        :param task_ids: The task IDs.
        :param task_graphs: The task graphs. Must be the same length as `task_ids`.
        :return: A list of tuples containing the parameters for each task input value. Each tuple
            contains:
            - Task ID.
            - Positional index of the input.
            - Type of the input.
            - Input value.
        """
        input_value_params = []
        for graph_index, task_graph in enumerate(task_graphs):
            for task_index, task in enumerate(task_graph.tasks):
                for position, task_input in enumerate(task.task_inputs):
                    if isinstance(task_input.value, core.TaskInputValue):
                        input_value_params.append(
                            (
                                task_ids[graph_index][task_index].bytes,
                                position,
                                task_input.type,
                                task_input.value,
                            )
                        )
        return input_value_params

    @staticmethod
    def _gen_task_input_output_ref_insertion_params(
        task_ids: Sequence[Sequence[UUID]],
        task_graphs: Sequence[core.TaskGraph],
    ) -> list[tuple[bytes, int, str, bytes, int]]:
        """
        Generates parameters for inserting task input output refs into the database.
        :param task_ids: The task IDs.
        :param task_graphs: The task graphs. Must be the same length as `task_ids`.
        :return: A list of tuples containing the parameters for each task input output ref. Each
            tuple contains:
            - Input task ID.
            - Positional index of the input.
            - Type of the input.
            - Output task ID.
            - Positional index of the output.
        """
        input_output_params = []
        for graph_index, task_graph in enumerate(task_graphs):
            for input_output_ref in task_graph.task_input_output_refs:
                input_output_params.append(
                    (
                        task_ids[graph_index][input_output_ref.input_task_index].bytes,
                        input_output_ref.input_position,
                        task_graph.tasks[input_output_ref.input_task_index]
                        .task_inputs[input_output_ref.input_position]
                        .type,
                        task_ids[graph_index][input_output_ref.output_task_index].bytes,
                        input_output_ref.output_position,
                    )
                )
        return input_output_params

    @staticmethod
    def _get_job_status(cursor: mariadb.Cursor, job: core.Job) -> core.JobStatus:
        """
        Gets the status of `job` from the database using the `cursor`.
        This method does not commit or rollback the transaction.
        :param cursor:
        :param job:
        :return: The job status.
        :raises StorageError: If the job is not found or if the job status is unknown.
        """
        cursor.execute(GetJobStatus, (job.job_id.bytes,))
        row = cursor.fetchone()
        if row is None:
            msg = f"No job found with id {job.job_id}."
            raise StorageError(msg)
        status_str = row[0]
        if status_str not in _StrToJobStatusMap:
            msg = f"Unknown job status: {status_str}."
            raise StorageError(msg)
        # Use fetchall after a fetchone to drain the result set even if it is already empty.
        cursor.fetchall()
        return _StrToJobStatusMap[status_str]

    @staticmethod
    def _get_data(cursor: mariadb.Cursor, data_id: core.DataId) -> core.Data:
        """
        Gets the data with `data_id` from the database using the `cursor`.
        This method does not commit or rollback the transaction.
        :param cursor:
        :param data_id:
        :return: The data.
        :raises StorageError: If the data is not found.
        """
        cursor.execute(GetData, (data_id.bytes,))
        row = cursor.fetchone()
        if row is None:
            msg = f"No data found with id {data_id}."
            raise StorageError(msg)
        value, hard_locality = row
        data = core.Data(id=data_id, value=value, hard_locality=hard_locality)
        cursor.execute(GetDataLocality, (data_id.bytes,))
        for (address,) in cursor.fetchall():
            data.localities.append(core.DataAddr(address))
        return data


def _raise_storage_error(message: str) -> None:
    """
    Raises a StorageError with the `message`.
    Workaround for ruff TRY301. See https://docs.astral.sh/ruff/rules/raise-within-try/.
    :param message:
    :raises StorageError: Always.
    """
    raise StorageError(message)



"""Custom type module for Spider."""

from __future__ import annotations

from typing import cast


class BoundedInt(int):
    """Bounded integer type."""

    def __new__(cls, value: int, bits: int = 32) -> BoundedInt:
        """Creates a bounded integer."""
        if bits not in (8, 16, 32, 64):
            msg = f"Unsupported bits size: {bits}. Supported sizes are 8, 16, 32, or 64."
            raise ValueError(msg)

        lower_bound = -(1 << (bits - 1))
        upper_bound = (1 << (bits - 1)) - 1

        if not (lower_bound <= value and value <= upper_bound):
            msg = (
                f"Bounded integer value ({value}) must be between {lower_bound} and {upper_bound}."
            )
            raise ValueError(msg)

        return super().__new__(cls, value)


class Int8(BoundedInt):
    """8 bits integer type."""

    def __new__(cls, value: int) -> Int8:
        """Creates an int8 integer."""
        return cast("Int8", super().__new__(cls, value, bits=8))


class Int16(BoundedInt):
    """16 bits integer type."""

    def __new__(cls, value: int) -> Int16:
        """Creates an int16 integer."""
        return cast("Int16", super().__new__(cls, value, bits=16))


class Int32(BoundedInt):
    """32 bits integer type."""

    def __new__(cls, value: int) -> Int32:
        """Creates an int32 integer."""
        return cast("Int32", super().__new__(cls, value, bits=32))


class Int64(BoundedInt):
    """64 bits integer type."""

    def __new__(cls, value: int) -> Int64:
        """Creates an int64 integer."""
        return cast("Int64", super().__new__(cls, value, bits=64))


class Float(float):
    """Float type."""

    def __new__(cls, value: float) -> Float:
        """Creates a float number."""
        return super().__new__(cls, value)


class Double(float):
    """Double type."""

    def __new__(cls, value: float) -> Double:
        """Creates a double number."""
        return super().__new__(cls, value)



"""Job module for Spider."""

from enum import IntEnum
from typing import TYPE_CHECKING
from uuid import UUID

if TYPE_CHECKING:
    from spider_py.core import TaskOutput

JobId = UUID


class JobStatus(IntEnum):
    """Job status."""

    Running = 0
    Succeeded = 1
    Failed = 2
    Cancelled = 3


class Job:
    """Represents a submitted job."""

    def __init__(self, job_id: JobId) -> None:
        """
        Initializes a running job.
        The job's status and results are cached from the data in storage. Once the job completes,
        the status and results will remain unchanged.
        :param job_id:
        """
        self.job_id = job_id
        self.status = JobStatus.Running
        self.results: list[TaskOutput] | None = None

    def is_running(self) -> bool:
        """:return: Whether the job is in `Running` status."""
        return self.status == JobStatus.Running



"""Data module for Spider."""

from dataclasses import dataclass, field
from uuid import UUID

DataId = UUID


DataAddr = str


@dataclass
class Data:
    """Represents a data object."""

    id: DataId
    value: bytes
    localities: list[DataAddr] = field(default_factory=list)
    hard_locality: bool = False
    persisted: bool = False



"""TaskGraph module for Spider."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from spider_py.core.task import Task


class TaskGraph:
    """
    Represents a task graph in Spider.
    TaskGraph represents a directed acyclic graph (DAG) of tasks.
    It stores:
    - tasks: A list of Task objects.
    - dependencies: A list of tuples representing the dependencies between tasks. Each tuple
      contains:
        - parent task index
        - child task index
    - input_task_indices: A list of task indices that have no parents (input tasks).
    - output_task_indices: A list of task indices that have no children (output tasks).
    - task_input_output_refs: A list of `InputOutputRef` representing references from task inputs
      to task outputs.
    """

    def __init__(self) -> None:
        """Initializes an empty task graph."""
        self.tasks: list[Task] = []
        self.dependencies: list[tuple[int, int]] = []
        self.input_task_indices: list[int] = []
        self.output_task_indices: list[int] = []
        self.task_input_output_refs: list[TaskGraph.InputOutputRef] = []

    class InputOutputRef:
        """Represents a reference from a task input to a task output."""

        def __init__(
            self,
            input_task_index: int,
            input_position: int,
            output_task_index: int,
            output_position: int,
        ) -> None:
            """Initializes an InputOutputRef object."""
            self.input_task_index: int = input_task_index
            self.input_position: int = input_position
            self.output_task_index: int = output_task_index
            self.output_position: int = output_position

        def with_offset(self, index_offset: int) -> TaskGraph.InputOutputRef:
            """
            Creates a new `InputOutputRef` with the task indices adjusted by applying the specified
            offset.

            :param index_offset: The offset to apply to the task indices.
            :return: A new `InputOutputRef` instance with adjusted task indices.
            """
            return TaskGraph.InputOutputRef(
                self.input_task_index + index_offset,
                self.input_position,
                self.output_task_index + index_offset,
                self.output_position,
            )

    def add_task(self, task: Task) -> None:
        """
        Adds a task to the graph.
        :param task: The task to add.
        """
        index = len(self.tasks)
        self.tasks.append(task)
        self.input_task_indices.append(index)
        self.output_task_indices.append(index)

    def copy(self) -> TaskGraph:
        """:return: A deep copy of the task graph."""
        return deepcopy(self)

    def merge_graph(self, graph: TaskGraph) -> None:
        """
        Merges another task graph into this task graph.
        :param graph: The task graph to merge.
        """
        index_offset = len(self.tasks)
        self.tasks.extend(graph.tasks)
        self.dependencies.extend(
            [
                (parent + index_offset, child + index_offset)
                for (parent, child) in graph.dependencies
            ]
        )
        self.input_task_indices.extend([index + index_offset for index in graph.input_task_indices])
        self.output_task_indices.extend(
            [index + index_offset for index in graph.output_task_indices]
        )
        self.task_input_output_refs.extend(
            [ref.with_offset(index_offset) for ref in graph.task_input_output_refs]
        )

    @staticmethod
    def chain_graph(parent: TaskGraph, child: TaskGraph) -> TaskGraph:
        """
        Chains two task graphs into a new task graph.
        :param parent: The parent task graph.
        :param child: The child task graph.
        :return: The chained task graph.
        :raise TypeError: If the outputs and the inputs of `graph` do not match.
        """
        graph = deepcopy(parent)
        index_offset = len(graph.tasks)
        parent_output_task_indices = graph.output_task_indices
        graph.tasks.extend(child.tasks)
        graph.dependencies.extend(
            [
                (parent_index + index_offset, child_index + index_offset)
                for (parent_index, child_index) in child.dependencies
            ]
        )
        graph.output_task_indices = [index + index_offset for index in child.output_task_indices]

        size_mismatch_msg = "Parent outputs size and child inputs size do not match."

        parent_output_task_it = iter(parent_output_task_indices)
        output_task_index = next(parent_output_task_it, None)
        output_position = 0
        for input_task_index in (i + index_offset for i in child.input_task_indices):
            input_task = graph.tasks[input_task_index]
            for input_position, task_input in enumerate(input_task.task_inputs):
                if output_task_index is None:
                    raise TypeError(size_mismatch_msg)

                if (output_task_index, input_task_index) not in graph.dependencies:
                    graph.dependencies.append((output_task_index, input_task_index))

                task_outputs = graph.tasks[output_task_index].task_outputs
                if output_position >= len(task_outputs):
                    raise TypeError(size_mismatch_msg)
                output_type = task_outputs[output_position].type
                input_type = task_input.type
                if input_type != output_type:
                    msg = f"Output type {output_type} does not match input type {input_type}"
                    raise TypeError(msg)

                graph.task_input_output_refs.append(
                    TaskGraph.InputOutputRef(
                        input_task_index, input_position, output_task_index, output_position
                    )
                )

                output_position += 1
                if output_position >= len(task_outputs):
                    output_position = 0
                    output_task_index = next(parent_output_task_it, None)

        if output_task_index is not None or output_position != 0:
            raise TypeError(size_mismatch_msg)

        graph.task_input_output_refs.extend(
            [ref.with_offset(index_offset) for ref in child.task_input_output_refs]
        )

        return graph



"""Executes a Spider Python task."""

from __future__ import annotations

import argparse
import inspect
import logging
from collections.abc import Sequence
from os import fdopen
from pydoc import locate
from types import FunctionType, GenericAlias
from typing import get_args, get_origin, get_type_hints, TYPE_CHECKING
from uuid import UUID

import msgpack

from spider_py import client
from spider_py.storage import MariaDBStorage, parse_jdbc_url, Storage
from spider_py.task_executor.task_executor_message import get_request_body, TaskExecutorResponseType
from spider_py.utils import from_serializable, to_serializable

if TYPE_CHECKING:
    from io import BufferedReader

# Set up logger
logger = logging.getLogger(__name__)


HeaderSize = 16


def parse_args() -> argparse.Namespace:
    """
    Parses task executor arguments.
    :return: The parsed arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--func", type=str, required=True, help="Name of the function to execute.")
    parser.add_argument(
        "--storage_url", type=str, required=True, help="JDBC URL for the storage backend."
    )
    parser.add_argument("--task_id", type=str, required=True, help="Task UUID.")
    parser.add_argument(
        "--input-pipe", type=int, required=True, help="File descriptor for the input pipe."
    )
    parser.add_argument(
        "--output-pipe", type=int, required=True, help="File descriptor for the output pipe."
    )
    return parser.parse_args()


def receive_message(pipe: BufferedReader) -> bytes:
    """
    Receives message from the pipe with a size header.
    :param pipe: Pipe to receive message from.
    :return: Received message body.
    :raises EOFError: If the message body size does not match header size.
    """
    body_size_str = pipe.read(HeaderSize).decode()
    body_size = int(body_size_str, base=10)
    body = pipe.read(body_size)
    if len(body) != body_size:
        msg = "Received message body size does not match the header size."
        raise EOFError(msg)
    return body


def parse_task_arguments(
    storage: Storage, params: list[inspect.Parameter], arguments: list[object]
) -> list[object]:
    """
    Parses arguments for the function to be executed.

    NOTE: `params` does not include the `TaskContext` parameter, and must be the same length as
    `arguments`. The caller is responsible for the size check.
    :param storage: Storage instance to use to get Data.
    :param params: A list of parameters in the function signature.
    :param arguments: A list of arguments to parse.
    :return: The parsed arguments.
    :raises TypeError: If a parameter has no type annotation or if an argument cannot be parsed.
    """
    parsed_args: list[object] = []
    for i, param in enumerate(params):
        arg = arguments[i]
        cls = param.annotation
        if param.annotation is inspect.Parameter.empty:
            msg = f"Parameter `{param.name}` has no type annotation."
            raise TypeError(msg)
        if cls is not client.Data:
            parsed_args.append(from_serializable(cls, arg))
            continue
        if not isinstance(arg, bytes):
            msg = f"Argument {i}: Expected `spider.Data` (bytes), but got {type(arg).__name__}."
            raise TypeError(msg)
        core_data = storage.get_data(UUID(bytes=arg))
        parsed_args.append(client.Data(core_data))
    return parsed_args


def parse_single_output_to_serializable(output: object, cls: type | GenericAlias) -> object:
    """
    Parses a single output from the function execution to a serializable form.
    :param output: Output to parse.
    :param cls: Expected output type.
    :return: The parsed output.
    """
    if isinstance(output, client.Data):
        return output.id.bytes
    return to_serializable(output, cls)


def parse_task_execution_results(
    results: object, types: type | GenericAlias | Sequence[type | GenericAlias]
) -> list[object]:
    """
    Parses results from the function execution.
    :param results: Results to parse.
    :param types: Expected output types. Must be a single type for non-tuple results, or a sequence
        of types matching the length of tuple results.
    :return: The parsed results.
    :raises TypeError: If the number of output types does not match the number of results.
    """
    response_messages: list[object] = [TaskExecutorResponseType.Result]
    if not isinstance(results, tuple):
        if not isinstance(types, (type, GenericAlias)):
            msg = "Invalid single output type."
            raise TypeError(msg)
        response_messages.append(parse_single_output_to_serializable(results, types))
        return response_messages
    # Parse as a tuple
    if not isinstance(types, Sequence) or len(results) != len(types):
        msg = "The number of output types does not match the number of results."
        raise TypeError(msg)
    for result, ret_type in zip(results, types, strict=True):
        response_messages.append(parse_single_output_to_serializable(result, ret_type))
    return response_messages


def get_return_types(
    func: FunctionType,
) -> type | GenericAlias | Sequence[type | GenericAlias]:
    """
    Gets the return types of a function.
    :param func: Function to get return types from.
    :return: Return types of the function. If the function returns a single value, the return type
        is a type or a generic alias. If the function returns multiple values, the return type is a
        sequence of types or generic aliases.
    :raises TypeError: If the function doesn't have return type annotation, or if the return type
        annotation is neither a type nor a generic alias.
    """
    signature = inspect.signature(func)
    annotation = signature.return_annotation

    if annotation is inspect.Signature.empty:
        msg = f"Function {func.__name__} has no return type annotation."
        raise TypeError(msg)

    # Resolve forward-referenced type annotations
    if isinstance(annotation, str):
        try:
            hints = get_type_hints(func)
            annotation = hints.get("return", annotation)
        except Exception as e:
            msg = f"Failed to get type hints for function {func.__name__}."
            raise TypeError(msg) from e

    origin = get_origin(annotation)
    if origin is not tuple:
        if not isinstance(annotation, (type, GenericAlias)):
            msg = (
                "Function return type annotation is neither a type nor a generic alias:"
                f" {annotation}."
            )
            raise TypeError(msg)
        return annotation
    return get_args(annotation)


def main() -> None:
    """Main function to execute the task."""
    # Parses arguments
    args = parse_args()
    function_name = args.func
    task_id = args.task_id
    task_id = UUID(task_id)
    storage_url = args.storage_url
    input_pipe_fd = args.input_pipe
    output_pipe_fd = args.output_pipe

    logger.debug("Function to run: %s", function_name)

    # Sets up storage
    storage_params = parse_jdbc_url(storage_url)
    storage = MariaDBStorage(storage_params)

    with fdopen(input_pipe_fd, "rb") as input_pipe, fdopen(output_pipe_fd, "wb") as output_pipe:
        input_message = receive_message(input_pipe)
        arguments = get_request_body(input_message)
        logger.debug("Args buffer parsed")

        # Get the function to run
        function = locate(function_name)
        if function is None or not inspect.isfunction(function):
            msg = f"{function_name} cannot be found in the current Python execution environment."
            raise ValueError(msg)

        signature = inspect.signature(function)
        if len(signature.parameters) != len(arguments) + 1:
            msg = (
                f"Function {function_name} expects {len(signature.parameters) - 1} arguments, but"
                f" {len(arguments)} were provided."
            )
            raise ValueError(msg)
        task_context = client.TaskContext(task_id, storage)
        arguments = [
            task_context,
            *parse_task_arguments(storage, list(signature.parameters.values())[1:], arguments),
        ]
        try:
            results = function(*arguments)
            logger.debug("Function %s executed", function_name)
            return_types = get_return_types(function)
            responses = parse_task_execution_results(results, return_types)
        except Exception as e:
            logger.exception("Function %s failed", function_name)
            responses = [
                TaskExecutorResponseType.Error,
                {"type": e.__class__.__name__, "message": str(e)},
            ]

        packed_responses = msgpack.packb(responses)
        output_pipe.write(f"{len(packed_responses):0{HeaderSize}d}".encode())
        output_pipe.write(packed_responses)
        output_pipe.flush()


if __name__ == "__main__":
    main()



"""Utility functions for job management."""

from spider_py import core

from .storage import Storage


def fetch_and_update_job_status(storage: Storage, job: core.Job) -> None:
    """
    Fetches and updates the job status from the storage.
    :param storage:
    :param job:
    """
    job.status = storage.get_job_status(job)


def fetch_and_update_job_results(storage: Storage, job: core.Job) -> None:
    """
    Fetches and updates the job status from the storage.

    NOTE: If the job results are already set, this method returns directly.

    :param storage:
    :param job:
    """
    if job.results is not None:
        return
    job.results = storage.get_job_results(job)



"""Spider type package."""

from spider_py.type.tdl_convert import to_tdl_type_str
from spider_py.type.tdl_parse import parse_tdl_type
from spider_py.type.type import Double, Float, Int8, Int16, Int32, Int64

__all__ = [
    "Double",
    "Float",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "parse_tdl_type",
    "to_tdl_type_str",
]



"""Storage backend interface module."""

from abc import ABC, abstractmethod
from collections.abc import Sequence

from spider_py import core
from spider_py.core import JobStatus


class StorageError(Exception):
    """Storage error."""

    def __init__(self, message: str) -> None:
        """Initializes storage error."""
        super().__init__(message)


class Storage(ABC):
    """Storage backend interface."""

    @abstractmethod
    def submit_jobs(
        self, driver_id: core.DriverId, task_graphs: Sequence[core.TaskGraph]
    ) -> Sequence[core.Job]:
        """
        Submits jobs to the storage.
        :param driver_id: Driver id.
        :param task_graphs: Task graphs to submit.
        :return: A list of submitted jobs.
        :raises StorageError: If the storage operations fail.
        """

    @abstractmethod
    def get_job_status(self, job: core.Job) -> JobStatus:
        """
        Gets the job status. This function does not set the `status` field in `job`.
        :param job:
        :return: The job status.
        :raises StorageError: If the storage operations fail.
        """

    @abstractmethod
    def get_job_results(self, job: core.Job) -> list[core.TaskOutput] | None:
        """
        Gets the job results. This function does not set the `results` field in `job`.
        :param job:
        :return: A list of task outputs if all tasks are finished.
        :return: None if any task output is not ready.
        :raises StorageError: If the storage operations fail.
        """

    @abstractmethod
    def create_data_with_driver_ref(self, driver_id: core.DriverId, data: core.Data) -> None:
        """
        Creates a data object in the storage with the given driver references to the data. This
        reference is used for garbage collection purposes.
        :param driver_id: The driver ID to associate with the data.
        :param data:
        :raises StorageError: If the storage operations fail.
        """

    @abstractmethod
    def create_data_with_task_ref(self, task_id: core.TaskId, data: core.Data) -> None:
        """
        Creates a data object in the storage with the given task references to the data. This
        reference is used for garbage collection purposes.
        :param task_id: The task ID to associate with the data.
        :param data:
        :raises StorageError: If the storage operations fail.

        """

    @abstractmethod
    def get_data(self, data_id: core.DataId) -> core.Data:
        """
        Gets the data object associated with the specified data ID from the storage.
        :param data_id:
        :return: The data object associated with `data_id`.
        :raises StorageError: If the storage operations fail.
        """

    @abstractmethod
    def create_driver(self, driver_id: core.DriverId) -> None:
        """
        Creates a driver in the storage.
        :param driver_id:
        :raises StorageError: If the storage operations fail.
        """



"""Spider core package."""

from .data import Data, DataAddr, DataId
from .driver import DriverId
from .job import Job, JobId, JobStatus
from .task import (
    Task,
    TaskId,
    TaskInput,
    TaskInputData,
    TaskInputOutput,
    TaskInputValue,
    TaskOutput,
    TaskOutputData,
    TaskOutputValue,
    TaskState,
)
from .task_graph import TaskGraph

__all__ = [
    "Data",
    "DataAddr",
    "DataId",
    "DriverId",
    "Job",
    "JobId",
    "JobStatus",
    "Task",
    "TaskGraph",
    "TaskId",
    "TaskInput",
    "TaskInputData",
    "TaskInputOutput",
    "TaskInputValue",
    "TaskOutput",
    "TaskOutputData",
    "TaskOutputValue",
    "TaskState",
]



"""Spider client driver module."""

from collections.abc import Sequence
from uuid import uuid4

import msgpack

from spider_py import core
from spider_py.client.data import Data
from spider_py.client.job import Job
from spider_py.client.task_graph import TaskGraph
from spider_py.storage import MariaDBStorage, parse_jdbc_url
from spider_py.type import parse_tdl_type
from spider_py.utils.serde import to_serializable

_argument_number_mismatch_msg = "Number of job inputs does not match number of arguments."


class Driver:
    """Spider client driver class."""

    def __init__(self, storage_url: str) -> None:
        """
        Creates a new Spider client driver and connects to the storage.
        :param storage_url: The URL of the storage to connect to.
        :raises StorageError: If the storage cannot be connected to.
        """
        self._driver_id = uuid4()
        self._storage = MariaDBStorage(parse_jdbc_url(storage_url))
        self._storage.create_driver(self._driver_id)

    def submit_jobs(
        self, task_graphs: Sequence[TaskGraph], args: Sequence[Sequence[object]]
    ) -> Sequence[Job]:
        """
        Submits a list of jobs to the storage.
        :param task_graphs: The list of task graphs to submit. Each task graph represents a job.
        :param args: The arguments for each job.
        :return: A sequence of `Job` objects representing the submitted jobs.
        :raises ValueError: If the number of job inputs does not match the number of arguments.
        """
        if len(task_graphs) != len(args):
            raise ValueError(_argument_number_mismatch_msg)

        if not task_graphs:
            return []

        core_task_graphs = []
        for task_graph, task_args in zip(task_graphs, args, strict=True):
            core_task_graphs.append(_copy_graph_for_submission(task_graph, task_args))

        core_jobs = self._storage.submit_jobs(self._driver_id, core_task_graphs)
        return [Job(core_job, self._storage) for core_job in core_jobs]

    def create_data(self, data: Data) -> None:
        """
        Creates a data in the storage.
        :param data:
        """
        self._storage.create_data_with_driver_ref(self._driver_id, data._impl)


def _copy_graph_for_submission(
    task_graph: TaskGraph, task_args: Sequence[object]
) -> core.TaskGraph:
    """
    Copies the given task graph for submission by setting task states and assigning input values.
    :param task_graph: The task graph to copy.
    :param task_args: The arguments to assign to the input tasks.
    :return: A copied core task graph ready for submission.
    """
    core_graph = task_graph._impl.copy()
    task_args_it = iter(task_args)
    next_arg = next(task_args_it, None)
    for task in core_graph.tasks:
        task.set_pending()
    for task_index in core_graph.input_task_indices:
        task = core_graph.tasks[task_index]
        task.set_ready()
        for task_input in task.task_inputs:
            if next_arg is None:
                raise ValueError(_argument_number_mismatch_msg)
            arg = next_arg
            next_arg = next(task_args_it, None)
            if isinstance(arg, Data):
                task_input.value = arg.id
                continue
            native_type = parse_tdl_type(task_input.type).native_type()
            serialized_value = to_serializable(arg, native_type)
            task_input.value = core.TaskInputValue(msgpack.packb(serialized_value))
    if next_arg is not None:
        raise ValueError(_argument_number_mismatch_msg)
    return core_graph



"""JDBC URL module."""

import urllib.parse
from dataclasses import dataclass


@dataclass
class JdbcParameters:
    """JDBC url parameters."""

    protocol: str
    host: str
    database: str
    port: int | None = None
    user: str | None = None
    password: str | None = None


_JdbcPrefix = "jdbc:"


def parse_jdbc_url(url: str) -> JdbcParameters:
    """
    Parses a JDBC URL.
    :param url:
    :return: The parsed JDBC parameters.
    :raises ValueError: If the JDBC URL is invalid.
    """
    protocol_prefix = ""
    if url.startswith(_JdbcPrefix):
        url = url.removeprefix(_JdbcPrefix)
        protocol_prefix = _JdbcPrefix
    parsed = urllib.parse.urlparse(url)

    msg = "Invalid JDBC URL: {}. Missing {}."
    if not parsed.scheme:
        raise ValueError(msg.format(url, "protocol"))
    if not parsed.hostname:
        raise ValueError(msg.format(url, "host"))
    if not parsed.path or not parsed.path.lstrip("/"):
        raise ValueError(msg.format(url, "database"))

    database = parsed.path.lstrip("/")
    query_params = urllib.parse.parse_qs(parsed.query)
    user = query_params.get("user", [None])[0]
    password = query_params.get("password", [None])[0]

    return JdbcParameters(
        protocol=protocol_prefix + parsed.scheme,
        host=parsed.hostname,
        port=parsed.port,
        database=database,
        user=user,
        password=password,
    )



"""Converts native types to TDL types."""

from collections.abc import Collection
from types import GenericAlias
from typing import get_args, get_origin

from spider_py.type.tdl_type import (
    BoolType,
    BytesType,
    ClassType,
    DoubleType,
    FloatType,
    Int8Type,
    Int16Type,
    Int32Type,
    Int64Type,
    ListType,
    MapType,
    TdlType,
)
from spider_py.type.type import Double, Float, Int8, Int16, Int32, Int64
from spider_py.type.utils import get_class_name

TypeDict = {
    Int8: Int8Type(),
    Int16: Int16Type(),
    Int32: Int32Type(),
    Int64: Int64Type(),
    Float: FloatType(),
    Double: DoubleType(),
    bool: BoolType(),
    bytes: BytesType(),
}


def _to_primitive_tdl_type(native_type: type | GenericAlias) -> TdlType | None:
    """
    Converts a native type to primitive TDL type.
    :param native_type:
    :return:
        - The converted TDL primitive if `native_type` is supported.
        - None if `native_type` is not a primitive Python type.
    :raises TypeError: If `native_type` is a primitive Python type not supported by TDL.
    """
    if isinstance(native_type, type) and native_type in TypeDict:
        return TypeDict[native_type]

    if native_type in (int, float, str, complex):
        msg = f"{native_type} is not a TDL type. Please use the corresponding TDL primitive type."
        raise TypeError(msg)

    return None


def to_tdl_type(native_type: type | GenericAlias) -> TdlType:
    """
    Converts a Python type to TDL type.
    :param native_type:
    :return: The converted TDL type.
    :raises TypeError: If `native_type` is not a valid TDL type.
    """
    primitive_tdl_type = _to_primitive_tdl_type(native_type)
    if primitive_tdl_type is not None:
        return primitive_tdl_type

    if isinstance(native_type, GenericAlias):
        origin = get_origin(native_type)
        if origin is list:
            args = get_args(native_type)
            if len(args) == 0:
                msg = "List does not have an element type."
                raise TypeError(msg)
            arg = args[0]
            return ListType(to_tdl_type(arg))

        if origin is dict:
            args = get_args(native_type)
            msg = "Dict does not have a key/value type."
            if len(args) != 2:  # noqa: PLR2004
                raise TypeError(msg)
            key = args[0]
            value = args[1]
            return MapType(to_tdl_type(key), to_tdl_type(value))

        msg = f"{native_type} is not a valid TDL type."
        raise TypeError(msg)

    if issubclass(native_type, Collection):
        msg = f"{native_type} is not a valid TDL type."
        raise TypeError(msg)

    return ClassType(get_class_name(native_type))


def to_tdl_type_str(native_type: type | GenericAlias) -> str:
    """
    :param native_type: A Python native type.
    :return: A string representation of the TDL type.
    :raises TypeError: If `native_type` is not a valid TDL type.
    """
    return to_tdl_type(native_type).type_str()



"""Spider client task context module."""

from spider_py import core
from spider_py.client.data import Data
from spider_py.storage import Storage


class TaskContext:
    """
    Represents the task context, providing:
    - Access to the task ID.
    - Task-referenced Data creation.
    """

    def __init__(self, task_id: core.TaskId, storage: Storage) -> None:
        """
        Initializes the task context.
        :param task_id:
        :param storage:
        """
        self._task_id = task_id
        self._storage = storage

    @property
    def task_id(self) -> core.TaskId:
        """:return: The task id."""
        return self._task_id

    def create_data(self, data: Data) -> None:
        """
        Creates a task-ID referenced data object in the storage.
        :param data:
        """
        self._storage.create_data_with_task_ref(self._task_id, data._impl)



"""Utility for TDL types."""

from importlib import import_module
from typing import cast


def get_class_name(cls: type) -> str:
    """
    :param cls:
    :return: Full name of `cls`.
    """
    return f"{cls.__module__}.{cls.__qualname__}"


def get_class_by_name(name: str) -> type:
    """
    Gets class by name.
    :param name:
    :return: The type object identified by `name`.
    :raise: TypeError if `class_name` is not a valid class.
    """
    module_name, _, class_name = name.rpartition(".")
    if module_name == "":
        msg = f"{name} does not contain a valid Python module."
        raise TypeError(msg)
    try:
        module = import_module(module_name)
        return cast("type", getattr(module, class_name))
    except (ValueError, ModuleNotFoundError, AttributeError) as e:
        msg = f"{name} is not a valid TDL type."
        raise TypeError(msg) from e



"""Parse TDL type string."""

from lark import Lark, LarkError, Token, Transformer, v_args

from spider_py.type.tdl_type import (
    BoolType,
    BytesType,
    ClassType,
    DoubleType,
    FloatType,
    Int8Type,
    Int16Type,
    Int32Type,
    Int64Type,
    ListType,
    MapType,
    TdlType,
)

grammar = r"""
type: map_type | list_type | base_type

map_type: "Map" "<" type "," type ">"
list_type: "List" "<" type ">"
base_type: ID

ID: /[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*/

%import common.WS
%ignore WS
"""

PrimitiveTypeDict = {
    "bool": BoolType(),
    "bytes": BytesType(),
    "double": DoubleType(),
    "float": FloatType(),
    "int8": Int8Type(),
    "int16": Int16Type(),
    "int32": Int32Type(),
    "int64": Int64Type(),
}


class TypeTransformer(Transformer[Token, TdlType]):
    """Transforms Lark parse tree into TDL type."""

    @v_args(inline=True)
    def type(self, value_type: TdlType) -> TdlType:
        """
        Unwraps the type node to return the TdlType.
        :param value_type: The type node to unwrap.
        :return: The unwrapped TdlType.
        """
        return value_type

    @v_args(inline=True)
    def map_type(self, key_type: TdlType, value_type: TdlType) -> TdlType:
        """
        Transforms map node into Map type.
        :param key_type: The key type of the map.
        :param value_type: The value type of the map.
        :return: The MapType with the given `key_type` and `value_type`.
        """
        return MapType(key_type, value_type)

    @v_args(inline=True)
    def list_type(self, element_type: TdlType) -> TdlType:
        """
        Transforms list node into List type.
        :param element_type: The element type of the list.
        :return: The ListType with the given `element_type`.
        """
        return ListType(element_type)

    @v_args(inline=True)
    def base_type(self, token_id: str) -> TdlType:
        """
        Transforms primitive node into primitive type.
        :param token_id: The token id of the primitive type.
        :return: The corresponding TdlType of the `token_id`.
        """
        if token_id in PrimitiveTypeDict:
            return PrimitiveTypeDict[token_id]
        return ClassType(token_id)


_parser = Lark(grammar, start="type", parser="lalr")
_transformer = TypeTransformer(visit_tokens=False)


def parse_tdl_type(type_string: str) -> TdlType:
    """
    Parses TDL type string into TDL type.
    :param type_string: TDL type string.
    :return: Parsed TDL type.
    :raises TypeError: If string is not a valid TDL type.
    """
    try:
        tree = _parser.parse(type_string)
        return _transformer.transform(tree)
    except LarkError as e:
        msg = f"Cannot parse TDL type `{type_string}`"
        raise TypeError(msg) from e



"""Spider python client."""

from .data import Data
from .driver import Driver
from .job import Job
from .task_context import TaskContext
from .task_graph import chain, group, TaskGraph

__all__ = [
    "Data",
    "Driver",
    "Job",
    "TaskContext",
    "TaskGraph",
    "chain",
    "group",
]



"""Spider Storage package."""

from .jdbc_url import JdbcParameters, parse_jdbc_url
from .job_utils import fetch_and_update_job_results, fetch_and_update_job_status
from .mariadb_storage import MariaDBStorage
from .storage import Storage, StorageError

__all__ = [
    "JdbcParameters",
    "MariaDBStorage",
    "Storage",
    "StorageError",
    "fetch_and_update_job_results",
    "fetch_and_update_job_status",
    "parse_jdbc_url",
]



"""Spider TDL types."""

from abc import ABC, abstractmethod
from types import GenericAlias

from typing_extensions import override

from spider_py.type.type import Double, Float, Int8, Int16, Int32, Int64
from spider_py.type.utils import get_class_by_name


class TdlType(ABC):
    """Abstract base class for all TDL types."""

    @abstractmethod
    def type_str(self) -> str:
        """:return: String representation of the TDL type."""

    @abstractmethod
    def native_type(self) -> type | GenericAlias:
        """:return: Native Python type of the TDL type."""


class DoubleType(TdlType):
    """TDL double type."""

    @override
    def type_str(self) -> str:
        return "double"

    @override
    def native_type(self) -> type | GenericAlias:
        return Double


class FloatType(TdlType):
    """TDL float type."""

    @override
    def type_str(self) -> str:
        return "float"

    @override
    def native_type(self) -> type | GenericAlias:
        return Float


class Int8Type(TdlType):
    """TDL int8 type."""

    @override
    def type_str(self) -> str:
        return "int8"

    @override
    def native_type(self) -> type | GenericAlias:
        return Int8


class Int16Type(TdlType):
    """TDL int16 type."""

    @override
    def type_str(self) -> str:
        return "int16"

    @override
    def native_type(self) -> type | GenericAlias:
        return Int16


class Int32Type(TdlType):
    """TDL int32 type."""

    @override
    def type_str(self) -> str:
        return "int32"

    @override
    def native_type(self) -> type | GenericAlias:
        return Int32


class Int64Type(TdlType):
    """TDL int64 type."""

    @override
    def type_str(self) -> str:
        return "int64"

    @override
    def native_type(self) -> type | GenericAlias:
        return Int64


class BoolType(TdlType):
    """TDL bool type."""

    @override
    def type_str(self) -> str:
        return "bool"

    @override
    def native_type(self) -> type | GenericAlias:
        return bool


class BytesType(TdlType):
    """TDL bytes type."""

    @override
    def type_str(self) -> str:
        return "bytes"

    @override
    def native_type(self) -> type | GenericAlias:
        return bytes


class ClassType(TdlType):
    """TDL Custom class type."""

    def __init__(self, name: str) -> None:
        """
        Creates a TDL custom class type.
        :param name: The name of the class.
        """
        self._name = name

    @override
    def type_str(self) -> str:
        return self._name

    @override
    def native_type(self) -> type | GenericAlias:
        """
        :return: Native Python type of the class.
        :raise: TypeError if `class_name` is not a valid class.
        """
        return get_class_by_name(self._name)


class ListType(TdlType):
    """TDL List type."""

    def __init__(self, element_type: TdlType) -> None:
        """
        Creates a TDL list type.
        :param element_type:
        """
        self.element_type = element_type

    @override
    def type_str(self) -> str:
        return f"List<{self.element_type.type_str()}>"

    @override
    def native_type(self) -> type | GenericAlias:
        return list[self.element_type.native_type()]  # type: ignore[misc]


def is_integral(tdl_type: TdlType) -> bool:
    """
    :param tdl_type:
    :return: Whether `tdl_type` is a TDL integral type.
    """
    return isinstance(tdl_type, (Int8Type, Int16Type, Int32Type, Int64Type))


def is_string(tdl_type: TdlType) -> bool:
    """
    :param tdl_type:
    :return: Whether `tdl_type` is a TDL string type, i.e. `List<int8>`.
    """
    return isinstance(tdl_type, ListType) and isinstance(tdl_type.element_type, Int8Type)


def is_map_key(tdl_type: TdlType) -> bool:
    """
    :param tdl_type:
    :return: Whether `tdl_type` is a supported key type of a map.
    """
    return is_integral(tdl_type) or is_string(tdl_type)


class MapType(TdlType):
    """TDL Map type."""

    def __init__(self, key_type: TdlType, value_type: TdlType) -> None:
        """
        Creates a TDL map type.
        :param key_type:
        :param value_type:
        :raises TypeError: If key is not a supported type.
        """
        if not is_map_key(key_type):
            msg = f"{key_type} is not a supported type for map key."
            raise TypeError(msg)
        self.key_type = key_type
        self.value_type = value_type

    @override
    def type_str(self) -> str:
        return f"Map<{self.key_type.type_str()},{self.value_type.type_str()}>"

    @override
    def native_type(self) -> type | GenericAlias:
        return dict[self.key_type.native_type(), self.value_type.native_type()]  # type: ignore[misc]



"""Spider client Data module."""

from __future__ import annotations

from uuid import uuid4

from spider_py import core


class Data:
    """Represents a spider client data."""

    def __init__(self, core_data: core.Data) -> None:
        """
        Initializes a data object by the given `core.Data` implementation.

        NOTE: This method is a low-level constructor. Please use `Data.from_value` to create a data
        object from a value.

        :param core_data:
        """
        self._impl = core_data

    @staticmethod
    def from_value(value: bytes) -> Data:
        """
        :param value: The data value.
        :return: A newly created data object with a random UUID and the given value.
        """
        return Data(core.Data(uuid4(), value))

    @property
    def id(self) -> core.DataId:
        """:return: The data id."""
        return self._impl.id

    @property
    def value(self) -> bytes:
        """:return: The data value."""
        return self._impl.value

    @property
    def localities(self) -> list[str]:
        """:return: The data localities."""
        return self._impl.localities

    @localities.setter
    def localities(self, value: list[str]) -> None:
        """
        Sets the data localities.
        :param value: The new localities.
        """
        self._impl.localities = value

    def add_locality(self, addr: str) -> None:
        """
        Adds a new address to the data localities.
        :param addr: The address to add.
        """
        self._impl.localities.append(addr)

    @property
    def hard_locality(self) -> bool:
        """:return: Whether the data has hard locality."""
        return self._impl.hard_locality

    @hard_locality.setter
    def hard_locality(self, hard: bool) -> None:
        """
        Sets whether the data has hard locality.
        :param hard:
        """
        self._impl.hard_locality = hard

    @property
    def persisted(self) -> bool:
        """:return: Whether the data is persisted."""
        return self._impl.persisted

    def set_persisted(self) -> None:
        """Sets the data as persisted."""
        self._impl.persisted = True
